#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "stree.h"


char docs_file[256];
char ksgs_file[256];

long s_parameter = 1;
long l_parameter = 2;
long h_parameter = 8000;
long b_parameter = 2;
double p_parameter = 1.0;
double q_parameter = 1.0;
const double EPS = 1e-8;

SUFFIX_TREE tree = NULL;
unsigned int* freq = NULL;
unsigned char* sel = NULL;
void print_help();
void process_args(int argc, char *argv[]);
int check_options();
int chomp_line(char* line);
void count_freq(STREE_NODE node);
void select_feature(STREE_NODE node);

int main(int argc, char* argv[])
{
	process_args(argc, argv);

    FILE* f_docs = fopen(docs_file, "r");
    if (! f_docs) {
        printf("Error: failed to open file %s.\n\n", docs_file);
        exit(1);
    }
    FILE* f_ksgs = fopen(ksgs_file, "w");
    if (! f_ksgs) {
        printf("Error: failed to open file %s.\n\n", ksgs_file);
        exit(1);
    }

    tree = stree_new_tree(STREE_ASCII, 1);

    char doc[MAXSTRLEN];
    unsigned int d_i = 0;
    while (fgets(doc, MAXSTRLEN, f_docs)) {
        int len = chomp_line(doc);
        int r = stree_add_string(tree, doc, len, d_i);
        if (r != 1) {
            printf("Error: failed to add the doc (id=%u)", d_i);
            if (r == 0) {
                printf("       becuse the program ran out of memory.\n\n");
            }
            if (r == 1) {
                printf("       becuse it contains non-ascii characters.\n\n");
            }
            exit(1);
        }
        d_i ++;
        if (d_i >= MAXNUMSTR) {
            printf("Error: the program can only process at most %u documents.\n\n", MAXNUMSTR);
            exit(1);
        }
    }

    int num_nodes = stree_get_num_nodes(tree);
    int i;
    freq = malloc(num_nodes * sizeof(unsigned int));
    if (freq == NULL) {
        printf("Error: the program ran out of memory.\n\n");
        exit(1);
    }
    for (i=0; i<num_nodes; i++) {
        freq[i] = 0;
    }
    sel = malloc(num_nodes * sizeof(unsigned char));
    if (sel == NULL) {
        printf("Error: the program ran out of memory.\n\n");
        exit(1);
    }
    for (i=0; i<num_nodes; i++) {
        sel[i] = 1;
    }

    STREE_NODE root = stree_get_root(tree);
    count_freq(root);
    select_feature(root);

    rewind(f_docs);
    while (fgets(doc, MAXSTRLEN, f_docs)) {
        int len = chomp_line(doc);
        STREE_NODE m_node;
        int m_pos;
        int m_len = stree_match(tree, doc, len, &m_node, &m_pos);
        assert(m_len == len);
        int i;
        for (i=0; i<len; i++) {
            int output = 1;
            if (s_parameter && (i > 0)) {
                if ((doc[i] == ' ') || (doc[i] == '\t')) {
                    output = 0;
                }
                if ((doc[i-1] != ' ') && (doc[i-1] != '\t')) {
                    output = 0;
                }
            }
            if (output) {
                STREE_NODE node = m_node;
                while (node != root) {
                    int n_i = stree_get_ident(tree, node);
                    if (sel[n_i]) {
                        fprintf(f_ksgs, "%x:%d ", n_i, freq[n_i]);
                    }
                    node = stree_get_parent(tree, node);
                }
            }
            m_node = stree_get_suffix_link(tree, m_node);
        }
        fprintf(f_ksgs, "\n");
    }

    // NOTE: the release of memory is ignored to save time.
    /*
    free(freq);
    freq = NULL;
    free(sel);
    sel = NULL;
    stree_delete_tree(tree);
    tree = NULL;
    */

    fclose(f_docs);
    fclose(f_ksgs);

	return 0;
}


void print_help()
{
    printf("Dell Zhang, Wee Sun Lee.\n");
    printf("Extracting Key-Substring-Group Features for Text Classification.\n");
    printf("In Proceedings of the 12th ACM SIGKDD International Conference on\n");
    printf("   Knowledge Discovery and Data Mining (KDD),\n");
    printf("Philadelphia, PA, Aug 2006.\n");
    printf("\n");
    printf("Usage: ksg [options] docs_file ksgs_file\n");
    printf("\n");
    printf("Options:\n");
    printf("         -?          -> help\n");
    printf("         -s [0,1]    -> assume white-spaces are word delimiters\n");
    printf("                        (default 1)\n");
    printf("         -l [2..]    -> the minimum frequency\n");
    printf("                        (default 2)\n");
    printf("         -h [l..]    -> the maximum frequency\n");
    printf("                        (default 8,000)\n");
    printf("         -b [2..]    -> the minimum number of branches\n");
    printf("                        (default 2)\n");
    printf("         -p (0..1]   -> the maximum parent-child conditional probability\n");
    printf("                        (default 1.0)\n");
    printf("         -q (0..1]   -> the maximum suffix-link conditional probability\n");
    printf("                        (default 1.0)\n");
    printf("Arguments:\n");
    printf("         docs_file    -> the input  file with each line as a raw document\n");
    printf("         ksgs_file    -> the output file with each line as a bag of ksg features\n");
}

void process_args(int argc,char *argv[])
{
    long i;
    for (i=1; ((i<argc) && ((argv[i])[0] == '-')); i++) {
        switch ((argv[i])[1]) {
        case '?':
            print_help();
            exit(0);
        case 's':
            i++;
            s_parameter = atol(argv[i]);
            break;
        case 'l':
            i++;
            l_parameter = atol(argv[i]);
            break;
        case 'h':
            i++;
            h_parameter = atol(argv[i]);
            break;
        case 'b':
            i++;
            b_parameter = atol(argv[i]);
            break;
        case 'p':
            i++;
            p_parameter = atof(argv[i]);
            break;
        case 'q':
            i++;
            q_parameter = atof(argv[i]);
            break;
        default:
            printf("Error: unrecognized option %s!\n\n",argv[i]);
            print_help();
            exit(1);
        }
    }
    if (! check_options()) {
        printf("Error: invalid option.\n\n");
        print_help();
        exit(1);
    }
    if((i+2) > argc) {
        // printf("Error: arguments are not enough.\n\n");
        print_help();
        exit(1);
    }
    strcpy(docs_file, argv[i+0]);
    strcpy(ksgs_file, argv[i+1]);
}

int check_options()
{
    if ((s_parameter != 0) && (s_parameter != 1)) {
        return 0;
    }
    if ((l_parameter < 2) || (h_parameter < l_parameter) || (b_parameter < 2)) {
        return 0;
    }
    if ((p_parameter <= 0) || (q_parameter <= 0)) {
        return 0;
    }
    return 1;
}

int chomp_line(char* line)
{
    int len = strlen(line);
    while ((line[len-1]=='\r') || (line[len-1]=='\n')) {
        line[len-1] = '\0';
        len --;
    }
    return len;
}

void count_freq(STREE_NODE node)
{
    int n_i = stree_get_ident(tree, node);
    freq[n_i] = stree_get_num_leaves(tree, node);
    STREE_NODE child = stree_get_children(tree, node);
    while (child != NULL) {
        count_freq(child);
        freq[n_i] += freq[stree_get_ident(tree, child)];
        child = stree_get_next(tree, child);
    }
}

void select_feature(STREE_NODE node)
{
    int n_i = stree_get_ident(tree, node);
    if (node == stree_get_root(tree)) {
        sel[n_i] = 0;
    }
    else {
        if ((freq[n_i] < l_parameter) || (freq[n_i] >= h_parameter)) {
            sel[n_i] = 0;
        }
        if (stree_get_num_children(tree, node) < b_parameter) {
            sel[n_i] = 0;
        }
        int p_i = stree_get_ident(tree, stree_get_parent(tree, node));
        if (freq[n_i]/freq[p_i] >= p_parameter+EPS) {
            sel[p_i] = 0;
        }
        int s_i = stree_get_ident(tree, stree_get_suffix_link(tree, node));
        if (freq[n_i]/freq[s_i] >= q_parameter+EPS) {
            sel[s_i] = 0;
        }
    }
    STREE_NODE child = stree_get_children(tree, node);
    while (child != NULL) {
        select_feature(child);
        child = stree_get_next(tree, child);
    }
}

